package jsr.daemon;

import java.io.*;
import java.util.*;

import jsr.daemon.logging.*;

/**
 * abstract generic daemon base class to be subclassed by daemons
 *
 * All singleton accesses are synchronized via this daemon instance if applicable.
 *
 * @author   Thomas Kopp
 * @version  0.9, 2000-11-16
 */
public abstract class GenericDaemon implements Daemon, DaemonConfig, ExtendedLogging {

    /**
     * initial capacity for attributes of this daemon
     */
    private int initialCapacity = 0;

    /**
     * name given to this daemon via its context
     */
    private String name = null;
    
    /**
     * logger set for this daemon via its context
     */
    private Logging logger = null;
    
    /**
     * resource for storing configuration information via a context
     */
    private DaemonAttributes attributes = null;

    /**
     * reference to the handler for dispatching listener events
     */
    private DaemonEventHandler handler = null;

    /**
     * reference to the service provider control interface of this daemon
     */
    private DaemonControl control = null;

    /**
     * reference to the service provider context interface of this daemon
     */
    private DaemonContext context = null;
    
    /**
     * reference to the application configuration interface of this daemon
     */
    private DaemonConfig config = null;
    
    /**
     * flag for keeping track of proper state
     */
    private boolean proper = false;
    
    /**
     * flag for keeping track of activity state
     */
    private boolean active = false;

    
    /** 
     * default constructor for this daemon
     */
    public GenericDaemon() {
        super();
    }

    /**
     * initialization method for this daemon
     * method is launched via the control interface and has to be subclassed
     * @exception DaemonException thrown by this daemon if initialization fails
     */
    public abstract void init() throws DaemonException;

    /**
     * destruction method for this daemon
     * method is launched via the control interface and has to be subclassed
     */
    public abstract void destroy();


    /**
     * access to the service provider context interface of this daemon
     * method is called by the container for supplying resources to be accessed via the configuration
     * @return service provider context interface of this daemon
     */
    protected DaemonContext getDaemonContext() {
        if (null == context) {
           synchronized(this) {            
               if (null == context) {
                  context = new DaemonContext() {
    
                                /**
                                 * method for naming this daemon
                                 * @param name the name given to a daemon
                                 */
                                public void setName(String name) {
                                    GenericDaemon.this.name = name;
                                }
    
                                /**
                                 * method for supplying a logging interface for this daemon 
                                 * @param logger logging resource to used by this daemon
                                 */
                                public void setLogger(Logging logger) {
                                    GenericDaemon.this.logger = logger;
                                }
    
                                /**
                                 * optional method for initializing the configuration interface
                                 * @param initialCapacity estimated number of attributes supplied
                                 */
                                public void setCapacity(int initialCapacity) {
                                    GenericDaemon.this.initialCapacity = initialCapacity;
                                }
    
                                /**
                                 * method for setting a daemon attribute
                                 * optionally provides for chronologically ordered attribute names
                                 * @param name the name of the attribute to be set
                                 * @param value the value of the attribute to be set
                                 */
                                public void setAttribute(String name, Object value) {
                                    getAttributes().put(name, value);
                                }
    
                                /**
                                 * method for removing a daemon attribute
                                 * @param name the name of the attribute to be removed
                                  */
                                public void removeAttribute(String name) {
                                    getAttributes().remove(name);
                                }
                            };
               }
           }
        }
        return context;
    }

    /**
     * access to the service provider control interface of this daemon
     * method is called by the container for controlling the life state of this daemon
     * @return service provider control interface of this daemon
     */
    protected DaemonControl getDaemonControl() {
        if (null == control) {
           synchronized(this) {
               if (null == control) {
                  control = new DaemonControl() {
    
                                /**
                                 * initialization method called in order to start the daemon
                                 * @exception DaemonException thrown by the daemon if initialization fails
                                 */
                                public void init() 
                                    throws DaemonException {
                                    if (!isActive()) {
                                       synchronized(this) {
                                           if (!isActive()) {
                                              proper = true;
                                              GenericDaemon.this.init();
                                              setActive(proper, null);
                                           }
                                       }
                                    }
                                }

                                /**
                                 * destruction method called in order to stop the daemon
                                 */
                                public void destroy() {
                                    if (isActive()) {
                                       synchronized(this) {
                                           if (isActive()) {
                                              GenericDaemon.this.destroy();
                                              setActive(false, null);
                                           }
                                       }
                                    }
                                }
                            };
               }
           }
        }
        return control;
    }
    
    /**
     * method for accessing the event handler of this daemon
     * @return event handler of this daemon
     */
    protected DaemonEventHandler getEventHandler() {
        if (null == handler) {
           synchronized(this) {
               if (null == handler) {
                  handler = new DaemonEventHandler();
               }
           }
        }
        return handler;
    }
    
    /**
     * method for accessing the attributes of this daemon
     * @return attribute resource of this daemon
     */
    private DaemonAttributes getAttributes() {
        if (null == attributes) {
           synchronized(this) {
               if (null == attributes) {
                  if (initialCapacity > 0) attributes = new DaemonAttributes(initialCapacity);
                  else attributes = new DaemonAttributes();
               }
           }
        }
        return attributes;
    }
    
    /**
     * method for managing the activity state of this daemon
     * @param active flag indicating the current activity state
     * @param throwable a throwable for indicating a possible exception
     */
    protected final void setActive(boolean active, Throwable throwable) {
        if (active != this.active) {
           DaemonEventHandler handler = getEventHandler();
           synchronized(handler) {
               if (active != this.active) {
                  this.active = active;
                  handler.active(this, active, throwable);
               }
           }
        }
    }

    /**
     * method for indicating the activity state of this daemon
     * used by this daemon for managing its own activity state
     * after calling down initialization is required for getting this daemon active again
     * @param throwable a throwable for indicating a possible exception
     */
    protected void down(Throwable throwable) {
        proper = false;
        setActive(false, throwable);
    }

    /**
     * life state indication of this daemon
     * state is managed by the container and/or by this daemon
     * @return true if and only if the implementing daemon is active
     */
    public boolean isActive() {
        return active;
    }
    
    /**
     * method for binding a listener to this daemon
     * @param listener the daemon listener to be bound
     */
    public void addListener(DaemonListener listener) {
        getEventHandler().addListener(listener);
    }

    /**
     * method for unbinding a listener from this daemon
     * @param listener the daemon listener to be unbound
     */
    public void removeListener(DaemonListener listener) {
        getEventHandler().removeListener(listener);
    }
    
    /**
     * descriptive information supplied by this daemon
     * the class name of this daemon is supplied as the default implementation
     * @return description of this daemon
     */
    public String getDaemonInfo() {
        return getClass().getName();
    }

    /**
     * finalizer method for this daemon
     */
    protected void finalize() throws Throwable {
        getDaemonControl().destroy();
        getEventHandler().removeListeners();
        handler = null;
        control = null;
        context = null;
        config = null;
        name = null;
        logger = null;
        attributes = null;
        super.finalize();
    }
    
    /**
     * access method for obtaining the application configuration interface of this daemon
     * interface methods are also available via this daemon class
     * @return configuration interface of this daemon
     */
    public DaemonConfig getDaemonConfig() {
        if (null == config) {
           synchronized(this) {
               if (null == config) {
                  config = new DaemonConfig() {
                       
                               /**
                                * method for accessing a configuration attribute of this daemon
                                * @param name the name of the attribute to be retrieved
                                * @return value of the attribute
                                */
                               public Object getAttribute(String name) { 
                                   return getAttributes().get(name);
                               }
 
                               /**
                                * method for listing attribute names of this configuration in chronological setting order
                                * using the enumeration should be synchronized for providing consistent behaviour
                                * @return enumeration of the attribute names available
                                */
                               public Enumeration getAttributeNames() {
                                   return getAttributes().keys();
                               }
                        
                               /**
                                * method for accessing the name of this daemon 
                                * @return name of this daemon set via the context
                                */
                               public String getDaemonName() {
                                   return name;
                               }
                           };
               }
           }
        }
        return config;
    }
    
    /**
     * method getAttribute() of the configuration interface
     * @param name the name of the attribute to be retrieved
     * @return value of the attribute
     */
    public Object getAttribute(String name) {
        return getDaemonConfig().getAttribute(name);
    }
    
    /**
     * method getAttributeNames() of the configuration interface
     * @return enumeration of the attribute names available
     */
    public Enumeration getAttributeNames() {
        return getDaemonConfig().getAttributeNames();
    }
    
    /**
     * method getDaemonName() of the configuration interface
     * @return name of this daemon set via the context
     */
    public String getDaemonName() {
        return getDaemonConfig().getDaemonName();
    }
    
    /**
     * method for logging output of this daemon
     * logging resources are supplied via the daemon context
     * @param message the message to be logged via an underlying logging resource
     */
    public void log(String message) {
        if ((null != message)&&(null != logger)) logger.log(message);
    }
    
    /**
     * method for logging extended output of this daemon
     * logging resources are supplied via the daemon context
     * @param message the message to be logged via an underlying logging resource
     * @param throwable additional throwable with its stack trace to be logged
     */
    public void log(String message, Throwable throwable) {
        if (null != logger) {
           if (logger instanceof ExtendedLogging) ((ExtendedLogging)logger).log(message, throwable);
           else {
                  logger.log(message);
                  if (null != throwable) {
                     StringWriter writer = new StringWriter();
                     PrintWriter printer = new PrintWriter(new BufferedWriter(writer));
                     throwable.printStackTrace(printer);
                     printer.close();
                     printer = null;
                     logger.log(writer.toString());
                  }
           }
        }
    }
}
